@@ -1,5 +1,6 @@
 /* -*- c-basic-offset: 2 -*- */
-/* Copyright(C) 2009-2016 Brazil
+/*
+  Copyright(C) 2009-2017 Brazil
 
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
@@ -358,11 +359,11 @@ grn_io_create(grn_ctx *ctx, const char *path, uint32_t header_size,
         GRN_MUNMAP(ctx, &grn_gctx, NULL, &fis->fmo, fis, header, b);
       }
       grn_fileinfo_close(ctx, fis);
-      if (grn_unlink(path) == -1) {
-        GRN_LOG(ctx, GRN_LOG_ERROR,
-                "failed to grn_unlink() path on grn_io_create() error: "
-                "<%s>: <%s>",
-                path, grn_strerror(errno));
+      if (grn_unlink(path) == 0) {
+        GRN_LOG(ctx, GRN_LOG_INFO,
+                "[io][create][error] removed path: <%s>", path);
+      } else {
+        ERRNO_ERR("[io][create][error] failed to remove path: <%s>", path);
       }
     }
     GRN_FREE(fis);
@@ -784,6 +785,14 @@ gen_pathname(const char *path, char *buffer, int fno)
   }
 }
 
+static uint32_t
+grn_io_n_files(grn_ctx *ctx, grn_io *io)
+{
+  unsigned long file_size;
+  file_size = grn_io_compute_file_size(io->header->version);
+  return ((io->header->curr_size + file_size - 1) / file_size);
+}
+
 grn_rc
 grn_io_size(grn_ctx *ctx, grn_io *io, uint64_t *size)
 {
@@ -791,16 +800,10 @@ grn_io_size(grn_ctx *ctx, grn_io *io, uint64_t *size)
   struct stat s;
   uint64_t tsize = 0;
   char buffer[PATH_MAX];
-  uint32_t nfiles;
+  uint32_t n_files;
 
-  if (io->header->curr_size) {
-    unsigned long file_size;
-    file_size = grn_io_compute_file_size(io->header->version);
-    nfiles = (uint32_t) ((io->header->curr_size + file_size - 1) / file_size);
-  } else {
-    nfiles = grn_io_max_n_files(io);
-  }
-  for (fno = 0; fno < nfiles; fno++) {
+  n_files = grn_io_n_files(ctx, io);
+  for (fno = 0; fno < n_files; fno++) {
     gen_pathname(io->path, buffer, fno);
     if (stat(buffer, &s)) {
       SERR("failed to stat path to compute size: <%s>",
@@ -821,10 +824,11 @@ grn_io_remove_raw(grn_ctx *ctx, const char *path)
   char buffer[PATH_MAX];
 
   if (grn_unlink(path) != 0) {
-    ERRNO_ERR("failed to remove path: <%s>",
+    ERRNO_ERR("[io][remove] failed to remove path: <%s>",
               path);
     return ctx->rc;
   }
+  GRN_LOG(ctx, GRN_LOG_INFO, "[io][remove] removed path: <%s>", path);
 
   for (fno = 1; ; fno++) {
     struct stat s;
@@ -832,9 +836,12 @@ grn_io_remove_raw(grn_ctx *ctx, const char *path)
     if (stat(buffer, &s) != 0) {
       break;
     }
-    if (grn_unlink(buffer) != 0) {
-      ERRNO_ERR("failed to remove path: <%s>",
-                buffer);
+    if (grn_unlink(buffer) == 0) {
+      GRN_LOG(ctx, GRN_LOG_INFO,
+              "[io][remove] removed numbered path: <%d>: <%s>", fno, buffer);
+    } else {
+      ERRNO_ERR("[io][remove] failed to remove numbered path: <%d>: <%s>",
+                fno, buffer);
       rc = ctx->rc;
     }
   }
@@ -1362,6 +1369,7 @@ grn_io_lock(grn_ctx *ctx, grn_io *io, int timeout)
 {
   static int _ncalls = 0, _ncolls = 0;
   uint32_t count, count_log_border = 1000;
+  uint32_t rc_check_interval = 1000;
   _ncalls++;
   if (!io) { return GRN_INVALID_ARGUMENT; }
   for (count = 0;; count++) {
@@ -1388,6 +1396,11 @@ grn_io_lock(grn_ctx *ctx, grn_io *io, int timeout)
                   "io(%s) collisions(%d/%d)", io->path, _ncolls, _ncalls);
         }
       }
+      if ((count % rc_check_interval) == 0) {
+        if (ctx->rc != GRN_SUCCESS) {
+          return ctx->rc;
+        }
+      }
       grn_nanosleep(GRN_LOCK_WAIT_TIME_NANOSECOND);
       continue;
     }
@@ -1446,19 +1459,36 @@ grn_io_flush(grn_ctx *ctx, grn_io *io)
     for (i = 0; i < max_mapped_segment; i++) {
       grn_io_mapinfo *info = &(io->maps[i]);
       uint32_t nth_file_info;
+      uint32_t *pnref;
+      uint32_t nref;
+      int msync_result;
 
       if (!info) {
         continue;
       }
+
+      pnref = &info->nref;
+      GRN_ATOMIC_ADD_EX(pnref, 1, nref);
+      if (nref != 0) {
+        GRN_ATOMIC_ADD_EX(pnref, -1, nref);
+        continue;
+      }
+
       if (!info->map) {
+        GRN_ATOMIC_ADD_EX(pnref, -1, nref);
+        GRN_FUTEX_WAKE(pnref);
         continue;
       }
 
       nth_file_info = grn_io_compute_nth_file_info(io, i);
-      if (GRN_MSYNC(ctx,
-                    io->fis[nth_file_info].fh,
-                    info->map,
-                    segment_size) != 0) {
+      msync_result = GRN_MSYNC(ctx,
+                               io->fis[nth_file_info].fh,
+                               info->map,
+                               segment_size);
+      GRN_ATOMIC_ADD_EX(pnref, -1, nref);
+      GRN_FUTEX_WAKE(pnref);
+
+      if (msync_result != 0) {
         rc = ctx->rc;
         break;
       }
@@ -1468,6 +1498,31 @@ grn_io_flush(grn_ctx *ctx, grn_io *io)
   return rc;
 }
 
+grn_bool
+grn_io_is_corrupt(grn_ctx *ctx, grn_io *io)
+{
+  uint32_t i;
+  uint32_t n_files;
+
+  if (!io) {
+    return GRN_FALSE;
+  }
+
+  n_files = grn_io_n_files(ctx, io);
+  for (i = 0; i < n_files; i++) {
+    char path[PATH_MAX];
+    struct stat s;
+    gen_pathname(io->path, path, i);
+    if (stat(path, &s) != 0) {
+      SERR("[io][corrupt] used path doesn't exist: <%s>",
+           path);
+      return GRN_TRUE;
+    }
+  }
+
+  return GRN_FALSE;
+}
+
 /** mmap abstraction **/
 
 static size_t mmap_size = 0;
@@ -1664,6 +1719,25 @@ grn_fileinfo_open_common(grn_ctx *ctx, fileinfo *fi, const char *path, int flags
       goto exit;
     }
 
+    switch (dwCreationDisposition) {
+    case CREATE_NEW :
+      GRN_LOG(ctx, GRN_LOG_INFO,
+              "[io][open] create new file: <%s>", path);
+      break;
+    case OPEN_ALWAYS :
+      if (GetLastError() == ERROR_ALREADY_EXISTS) {
+        GRN_LOG(ctx, GRN_LOG_INFO,
+                "[io][open] open existing file because it exists: <%s>", path);
+      } else {
+        GRN_LOG(ctx, GRN_LOG_INFO,
+                "[io][open] create new file because it doesn't exist: <%s>",
+                path);
+      }
+      break;
+    default :
+      break;
+    }
+
     if (grn_io_use_sparse) {
       FILE_SET_SPARSE_BUFFER buffer;
       buffer.SetSparse = TRUE;
@@ -1699,6 +1773,8 @@ grn_fileinfo_open_common(grn_ctx *ctx, fileinfo *fi, const char *path, int flags
            path);
       goto exit;
     }
+    GRN_LOG(ctx, GRN_LOG_INFO,
+            "[io][open] truncated: <%s>", path);
     goto exit;
   }
   /* O_RDWR only */
@@ -1711,6 +1787,8 @@ grn_fileinfo_open_common(grn_ctx *ctx, fileinfo *fi, const char *path, int flags
          path);
     goto exit;
   }
+  GRN_LOG(ctx, GRN_LOG_INFO,
+          "[io][open] open existing file: <%s>", path);
 
 exit :
   return ctx->rc;
@@ -1721,8 +1799,7 @@ grn_fileinfo_open(grn_ctx *ctx, fileinfo *fi, const char *path, int flags)
 {
   grn_rc rc;
   struct _grn_io_header io_header;
-  DWORD header_size;
-  DWORD read_bytes;
+  LARGE_INTEGER file_size;
   int version = grn_io_version_default;
 
   rc = grn_fileinfo_open_common(ctx, fi, path, flags);
@@ -1730,7 +1807,9 @@ grn_fileinfo_open(grn_ctx *ctx, fileinfo *fi, const char *path, int flags)
     return rc;
   }
 
-  if (!(flags & O_CREAT)) {
+  if (GetFileSizeEx(fi->fh, &file_size) && file_size.QuadPart > 0) {
+    DWORD header_size;
+    DWORD read_bytes;
     header_size = sizeof(struct _grn_io_header);
     ReadFile(fi->fh, &io_header, header_size, &read_bytes, NULL);
     if (read_bytes == header_size) {
